import abc
import re
import os
import sys
from enum import Enum
from itertools import groupby

from advarchs.utils import run_with_output


CONSOLE_CODING = 'utf8'
if sys.platform == 'win32':
    CONSOLE_CODING = 'cp866'


class ArchiveStatus(Enum):
    ALL_GOOD = 0
    NOT_COMPATIBLE = 1
    CORRUPTED = 2
    UNKNOWN_ERROR = 3


class AdvarchsExtractException(Exception):
    pass


class ArchiveHandlerBase(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def unpacker(self):
        """"""
        return

    @abc.abstractmethod
    def files_list(self, apath):
        """ Get the archive's content list """
        return

    @abc.abstractmethod
    def check(self, apath):
        """ Check archive if it's corrupted """
        return

    @abc.abstractmethod
    def extract(self, apath, ffilter):
        """ Extract files from archive """
        return


class HandlersFactory:
    """ Factory to get """
    _handlers = {}

    @classmethod
    def get_handler(cls, name):
        try:
            return cls._handlers[name]
        except KeyError:
            raise ValueError(name)

    @classmethod
    def register(cls, name, handler):
        cls._handlers[name] = handler

    @classmethod
    def handlers(cls):
        return cls._handlers


class SevenZHandler(ArchiveHandlerBase):

    re_not_compatible_pattern = r"can not open the file as archive"
    re_success_pattern = r"everything is ok"
    re_corruption_pattern = r"data error"

    def __init__(self, unpacker):
        self._unpacker = unpacker

    def _files_info(self, apath):
        r, out, _ = run_with_output([self.unpacker(), 'l', '-slt', apath], CONSOLE_CODING)
        raw = re.sub(r"^-+", '-' * 10, out, flags=re.MULTILINE).split('-' * 10)[2]
        blocks = []
        current_block = {}
        # make flat list of clean stdout's lines
        lines = [line.strip() for line in raw.strip().split('\n')]
        # break down flat list with details on lists for each file in archive
        block_list = [list(g) for k, g in groupby(lines, lambda x: x != '') if k]

        for block in block_list:
            for line in block:
                key, _, value = line.partition('=')
                current_block[key.strip().lower()] = value.strip()
            blocks.append(current_block)
            current_block = {}

        return blocks

    def _tech_info(self, apath):
        r, out, _ = run_with_output([self.unpacker(), 'l', '-slt', apath], CONSOLE_CODING)
        raw = re.sub(r"^-+", '-' * 10, out, flags=re.MULTILINE).split('-' * 10)[1]
        t_info = {}
        for line in raw.strip().split('\n'):
            key, _, value = line.partition('=')
            t_info[key.strip().lower()] = value.strip().lower()

        return t_info

    def unpacker(self):
        return self._unpacker

    def check(self, apath):

        r, out, err = run_with_output([self.unpacker(), 't', apath], CONSOLE_CODING)
        if r == 0:
            return ArchiveStatus.ALL_GOOD
        else:
            if not re.search(self.re_success_pattern, out):
                if re.search(self.re_not_compatible_pattern, err):
                    return ArchiveStatus.NOT_COMPATIBLE
                elif re.search(self.re_corruption_pattern, err):
                    return ArchiveStatus.CORRUPTED
                else:
                    return ArchiveStatus.UNKNOWN_ERROR

    def files_list(self, apath):
        a_files_info = self._files_info(apath)
        return [f["path"] for f in a_files_info]

    def extract(self, apath, afile):
        dfolder = '-o' + os.path.dirname(apath)
        args = [self.unpacker(), 'e', '-aoa', apath, afile, dfolder]
        r, out, err = run_with_output(args, CONSOLE_CODING)
        if r != 0:
            raise AdvarchsExtractException('Could not extract archive.')
        f_path = os.path.join(os.path.dirname(apath), afile)

        return f_path


class RarHandler(ArchiveHandlerBase):

    re_not_compatible_pattern = r"is not rar archive"
    re_success_pattern = r"all ок"
    re_corruption_pattern = r"total errors:"

    def __init__(self, unpacker):
        self._unpacker = unpacker

    def unpacker(self):
        return self._unpacker

    def _files_info(self, apath):
        r, out, _ = run_with_output([self.unpacker(), 'ltab', apath], CONSOLE_CODING)
        raw = re.search(r"([ \t]*name:.*?)(?=(\n{1,2}service:\s*eof|\Z))", out, re.M | re.S).group(1)
        blocks = []
        current_block = {}
        # make flat list of clean stdout's lines
        lines = [line.strip() for line in raw.strip().split('\n')]
        # break down flat list with details on lists for each file in archive
        block_list = [list(g) for k, g in groupby(lines, lambda x: x != '') if k]

        for block in block_list:
            for line in block:
                key, _, value = line.partition(': ')
                current_block[key.strip().lower()] = value.strip()
            blocks.append(current_block)
            current_block = {}

        return blocks

    def files_list(self, apath):
        a_files_info = self._files_info(apath)
        return [f["name"] for f in a_files_info if "type" in f.keys() and f["type"].lower() == "file"]

    def check(self, apath):
        r, out, err = run_with_output([self.unpacker(), 't', apath], CONSOLE_CODING)
        if r == 0:
            return ArchiveStatus.ALL_GOOD
        else:
            if not re.search(self.re_success_pattern, out):
                if re.search(self.re_not_compatible_pattern, out):
                    return ArchiveStatus.NOT_COMPATIBLE
                elif re.search(self.re_corruption_pattern, out):
                    return ArchiveStatus.CORRUPTED
                else:
                    return ArchiveStatus.UNKNOWN_ERROR

    def extract(self, apath, afile):
        args = [self.unpacker(), 'e', '-o+', apath, afile]
        r, out, err = run_with_output(args, CONSOLE_CODING, cwd=os.path.dirname(apath))
        if r != 0:
            raise AdvarchsExtractException('Could not extract archive.')
        f_path = os.path.join(os.path.dirname(apath), afile)

        return f_path



import os
from shutil import which
import string

from mailbox import Message

import requests

from advarchs.utils import file_ext, extract_version
from advarchs.handlers import (HandlersFactory, AdvarchsExtractException,
                               ArchiveStatus, SevenZHandler, RarHandler)


__all__ = ['webfilename', 'extract_web_archive']


ARCHIVE_COMPRESS_FORMATS = ('rar', 'tar', 'zip', 'gzip', 'bzip', 'gz')

__version__ = extract_version(__file__)


REMOVE_PUNCTUATION = dict((ord(char), None) for char in string.punctuation if char not in '._-')

SEVENZ = max([a if which(a) else None for a in ('7z', '7za')], key=lambda x: bool(x), default=None)
UNRAR = 'unrar' if which('unrar') else None

if os.name == 'nt':
    if SEVENZ:
        HandlersFactory.register(SEVENZ, SevenZHandler(SEVENZ))
    else:
        raise AdvarchsExtractException('Unpacker is not installed.')
elif os.name == 'posix':
    if SEVENZ and UNRAR:
        HandlersFactory.register(SEVENZ, SevenZHandler(SEVENZ))
        HandlersFactory.register(UNRAR, RarHandler(UNRAR))
    elif SEVENZ and (not UNRAR):
        HandlersFactory.register(SEVENZ, SevenZHandler(SEVENZ))
    else:
        raise AdvarchsExtractException('Unpacker is not installed.')


class AdvarchsDownloadException(Exception):
    pass


def filename_from_content_disposition(content_disposition):
    """
    Extract and validate filename from a Content-Disposition header.
    """
    msg = Message('Content-Disposition: %s' % content_disposition)
    filename = msg.get_filename()
    if filename:
        # Basic sanitation.
        filename = os.path.basename(filename).lstrip('.').strip()
        if filename:
            return filename


def _get_headers_from_url(url):
    with requests.get(url, stream=True) as r:
        return r.headers


def _content_disposition(headers):
    return headers.get("content-disposition")


def webfilename(url):
    """Get filename from archive's headers or from the url"""
    headers = _get_headers_from_url(url)

    if _content_disposition(headers):
        result = filename_from_content_disposition(_content_disposition(headers))
    else:
        result = url.split("/")[-1]

    return result.translate(REMOVE_PUNCTUATION)


def download(url, fpath):
    """Download file using streaming"""
    with open(fpath, 'wb') as f:
        try:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                f_size = 0
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
                        f_size += len(chunk)
            return f_size
        except Exception as e:
            os.remove(fpath)
            raise AdvarchsDownloadException('Could not download file.' + e.message)


def _is_matched(afile, ffilter=[]):
    """ Check if file in archive should be extracted """
    if ffilter:
        # we check only full name and extension of file
        if (os.path.basename(afile) in ffilter) or (file_ext(afile) in ffilter):
            return True
        else:
            return False
    return True


def _is_archive(afile):
    """Check if inside file can be archive or compressed"""
    return file_ext(os.path.basename(afile)) in ARCHIVE_COMPRESS_FORMATS


def resolve_format(apath):
    """ Resolve right handler for archive """
    status = ArchiveStatus.NOT_COMPATIBLE
    handlers = HandlersFactory.handlers()
    for handler in dict(handlers):
        unpacker = handlers[handler]
        status = unpacker.check(apath)
        if unpacker.check(apath) == ArchiveStatus.ALL_GOOD:
            return handler
    if status == ArchiveStatus.CORRUPTED:
        raise AdvarchsExtractException('Could not unpack. Archive is corrupted')
    elif status == ArchiveStatus.NOT_COMPATIBLE:
        raise AdvarchsExtractException('Could not unpack. Archive format is not supported')
    elif status == ArchiveStatus.UNKNOWN_ERROR:
        raise AdvarchsExtractException('Could not unpack. Unknown error')


def extract(apath, ffilter=[]):
    """Extract all or specified files from archive"""
    _extracted_files = []
    _is_nested = False

    def extract_recursive(curr_apath):
        """ Extract while archives or compression occur """
        handler = resolve_format(curr_apath)
        unpacker = HandlersFactory.get_handler(handler)
        files = unpacker.files_list(curr_apath)

        for f in files:
            if _is_matched(f, ffilter=ffilter):
                _fpath = unpacker.extract(curr_apath, f)
                _extracted_files.append(_fpath)
            if _is_archive(f):
                _apath = unpacker.extract(curr_apath, f)
                extract_recursive(_apath)

    extract_recursive(apath)
    return _extracted_files


def extract_web_archive(url, apath, ffilter=[]):
    """ Download archive and extract all or specified files"""
    download(url, apath)
    output_files = extract(apath, ffilter=ffilter)
    return output_files


print('asdsa')


import os
import hashlib
import tarfile
import subprocess as subp
from random import choice, randint
from string import ascii_uppercase
from zipfile import ZipFile
from pathlib import Path

import tomlkit


def extract_version(source_file):
    """ Extract package version from pyproject file """
    d = Path(source_file)
    result = None
    while d.parent != d and result is None:
        d = d.parent
        pyproject_toml_path = d / 'pyproject.toml'
        if pyproject_toml_path.exists():
            with open(file=str(pyproject_toml_path)) as f:
                pyproject_toml = tomlkit.parse(string=f.read())
                if 'tool' in pyproject_toml and 'poetry' in pyproject_toml['tool']:
                    # noinspection PyUnresolvedReferences
                    result = pyproject_toml['tool']['poetry']['version']
    return result


def get_hash_memory_optimized(f_path, mode='sha256'):
    """ Get hash of file"""
    h = hashlib.new(mode)
    with open(f_path, 'rb') as file:
        block = file.read(4096)
        while block:
            h.update(block)
            block = file.read(4096)

    return h.hexdigest()


def read_file(file, encoding="utf8"):
    """ Simple reading text file """
    with open(file, "r", encoding=encoding) as f:
        result = f.read()
    return result


def create_file(file, data, encoding="utf8"):
    """ Create text file and return hash of it """
    with open(file, "w", encoding=encoding) as f:
        f.write(data)
    return get_hash_memory_optimized(file)


def add_file_to_zip(fpath, apath):
    """ Add file to zip archive"""
    with ZipFile(apath, 'a') as zip_o:
        zip_o.write(fpath, arcname=os.path.basename(fpath))


def add_file_to_tar(fpath, apath):
    """ Add file to tar archive"""
    with tarfile.open(apath, "a") as tar_o:
        tar_o.add(fpath, arcname=os.path.basename(fpath))


def file_ext(f):
    """ File extension """
    return Path(f).suffix.replace('.', '')


def random_string(min_l, max_l):
    """ Get random text of length ranging from min_l to max_l """
    return ''.join(choice(ascii_uppercase) for i in range(randint(min_l, max_l)))


def run_with_output(args, encoding, **kwargs):
    """ Run program """
    res = subp.Popen(args, stdout=subp.PIPE, stderr=subp.PIPE, **kwargs)
    stdout, stderr = res.communicate()
    return res.returncode, stdout.decode(encoding).lower(), stderr.decode(encoding).lower()




# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['advarchs']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.22,<3.0', 'tomlkit>=0.5.5,<0.6.0']

setup_kwargs = {
    'name': 'advarchs',
    'version': '0.1.6',
    'description': 'Data retrieval from remote archives',
    'long_description': 'Advarchs: Data retrieval from remote archives\n=============================================\n\n.. image:: https://img.shields.io/pypi/v/advarchs.svg\n   :target: https://pypi.python.org/pypi/advarchs\n   :alt: PyPI Version\n\n.. image:: https://img.shields.io/pypi/pyversions/advarchs.svg\n   :target: https://pypi.python.org/pypi/advarchs\n   :alt: Supported Python Versions\n\n.. image:: https://img.shields.io/travis/elessarelfstone/advarchs/master.svg\n   :target: https://travis-ci.org/elessarelfstone/advarchs\n   :alt: Build Status\n\n.. image:: https://img.shields.io/badge/wheel-yes-brightgreen.svg\n   :target: https://pypi.python.org/pypi/advarchs\n   :alt: Wheel Status\n\nOverview\n--------\nAdvarchs is simple tool for retrieving data from web archives.\nIt is especially useful if you are working with remote data stored in compressed\nspreadsheets or of similar format.\n\nGetting Started\n---------------\n\nSay you need to perform some data anlytics on an excel spreadsheet that gets\nrefreshed every month and stored in RAR format. You can target a that file\nand convert it to a pandas_ dataframe with the following procedure:\n\n.. code-block:: python\n\n    import pd\n    import os\n    import tempfile\n    from advarchs import webfilename,extract_web_archive\n\n    TEMP_DIR = tempfile.gettempdir()\n\n    url = "http://www.site.com/archive.rar"\n    arch_file_name = webfilename(url)\n    arch_path = os.path.join(TEMP_DIR, arch_file_name)\n    xlsx_files = extract_web_archive(url, arch_path, ffilter=[\'xlsx\'])\n    for xlsx_f in xlsx_files:\n        xlsx = pd.ExcelFile(xlspath)\n\n    ...\n\nRequirements\n------------\n\n- ``Python 3.5+``\n- ``p7zip``\n\nSpecial note\n~~~~~~~~~~~~\n\nOn CentOS and Ubuntu <= 16.04, the following packages are needed:\n\n- ``unrar``\n\nInstallation\n------------\n\n.. code-block:: shell\n\n    pip install advarchs\n\nContributing\n------------\nSee `CONTRIBUTING`_\n\nCode of Conduct\n~~~~~~~~~~~~~~~\nThis project adheres to the `Contributor Covenant 1.2`_.\nBy participating, you are advised to adhere to this Code of Conduct in all your\ninteractions with this project.\n\nLicense\n-------\n\n`Apache-2.0`_\n\n.. _`pandas`: https://pypi.org/project/pandas/\n.. _`CONTRIBUTING`: https://github.com/elessarelfstone/advarchs/blob/master/CONTRIBUTING.md\n.. _`Contributor Covenant 1.2`: http://contributor-covenant.org/version/1/2/0\n.. _`Apache-2.0`: https://github.com/elessarelfstone/advarchs/blob/master/LICENSE\n',
    'author': 'Dauren Sdykov',
    'author_email': 'elessarelfstone@mail.ru',
    'url': 'https://github.com/elessarelfstone/advarchs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)
